@@ -4,7 +4,7 @@ use strict;
 use warnings;
 use Config;
 
-our $VERSION = '2.000_009'; # 2.0.9
+our $VERSION = '2.000_016';
 $VERSION = eval $VERSION;
 
 BEGIN {
@@ -21,10 +21,25 @@ our $_ROOT = _WIN32 ? do {
   my $UNC = qr{[\\/]{2}[^\\/]+[\\/][^\\/]+};
   qr{^(?:$UNC|[A-Za-z]:|)$_DIR_SPLIT};
 } : qr{^/};
-our ($_PERL) = $^X =~ /(.+)/; # $^X is internal how could it be tainted?!
+our $_PERL;
 
 sub _cwd {
   my $drive = shift;
+  if (!$_PERL) {
+    ($_PERL) = $^X =~ /(.+)/; # $^X is internal how could it be tainted?!
+    if (_is_abs($_PERL)) {
+    }
+    elsif (-x $Config{perlpath}) {
+      $_PERL = $Config{perlpath};
+    }
+    else {
+      ($_PERL) =
+        map { /(.*)/ }
+        grep { -x $_ }
+        map { join($_DIR_JOIN, $_, $_PERL) }
+        split /\Q$Config{path_sep}\E/, $ENV{PATH};
+    }
+  }
   local @ENV{qw(PATH IFS CDPATH ENV BASH_ENV)};
   my $cmd = $drive ? "eval { Cwd::getdcwd(q($drive)) }"
                    : 'getcwd';
@@ -49,12 +64,20 @@ sub _catdir {
   }
 }
 
+sub _is_abs {
+  if (_USE_FSPEC) {
+    require File::Spec;
+    File::Spec->file_name_is_absolute($_[0]);
+  }
+  else {
+    $_[0] =~ $_ROOT;
+  }
+}
+
 sub _rel2abs {
   my ($dir, $base) = @_;
   return $dir
-    if (_USE_FSPEC && require File::Spec)
-      ? File::Spec->file_name_is_absolute($dir)
-      : $dir =~ $_ROOT;
+    if _is_abs($dir);
 
   $base = _WIN32 && $dir =~ s/^([A-Za-z]:)// ? _cwd("$1")
         : $base                              ? $base
@@ -69,6 +92,7 @@ sub import {
 
   my @steps;
   my %opts;
+  my %attr;
   my $shelltype;
 
   while (@args) {
@@ -111,18 +135,21 @@ DEATH
     elsif ( $arg eq '--no-create' ) {
       $opts{no_create} = 1;
     }
+    elsif ( $arg eq '--quiet' ) {
+      $attr{quiet} = 1;
+    }
     elsif ( $arg =~ /^--/ ) {
       die "Unknown import argument: $arg";
     }
     else {
-      push @steps, ['activate', $arg];
+      push @steps, ['activate', $arg, \%opts];
     }
   }
   if (!@steps) {
-    push @steps, ['activate', undef];
+    push @steps, ['activate', undef, \%opts];
   }
 
-  my $self = $class->new(%opts);
+  my $self = $class->new(%attr);
 
   for (@steps) {
     my ($method, @args) = @$_;
@@ -153,7 +180,7 @@ sub libs { $_[0]->{libs}   ||= [ \'PERL5LIB' ] }
 sub bins { $_[0]->{bins}   ||= [ \'PATH' ] }
 sub roots { $_[0]->{roots} ||= [ \'PERL_LOCAL_LIB_ROOT' ] }
 sub extra { $_[0]->{extra} ||= {} }
-sub no_create { $_[0]->{no_create} }
+sub quiet { $_[0]->{quiet} }
 
 my $_archname = $Config{archname};
 my $_version  = $Config{version};
@@ -206,9 +233,9 @@ sub lib_paths_for {
 
 sub _mm_escape_path {
   my $path = shift;
-  $path =~ s/\\/\\\\\\\\/g;
+  $path =~ s/\\/\\\\/g;
   if ($path =~ s/ /\\ /g) {
-    $path = qq{"\\"$path\\""};
+    $path = qq{"$path"};
   }
   return $path;
 }
@@ -294,11 +321,12 @@ sub deactivate_all {
 }
 
 sub activate {
-  my ($self, $path) = @_;
+  my ($self, $path, $opts) = @_;
+  $opts ||= {};
   $self = $self->new unless ref $self;
   $path = $self->resolve_path($path);
-  $self->ensure_dir_structure_for($path)
-    unless $self->no_create;
+  $self->ensure_dir_structure_for($path, { quiet => $self->quiet })
+    unless $opts->{no_create};
 
   $path = $self->normalize_path($path);
 
@@ -309,7 +337,7 @@ sub activate {
   }
 
   my %args;
-  if (!@active_lls || $active_lls[0] ne $path) {
+  if ($opts->{always} || !@active_lls || $active_lls[0] ne $path) {
     %args = (
       bins  => [ $self->install_base_bin_path($path), @{$self->bins} ],
       libs  => [ $self->install_base_perl_path($path), @{$self->libs} ],
@@ -331,7 +359,19 @@ sub normalize_path {
 }
 
 sub build_environment_vars_for {
-  my $self = $_[0]->new->activate($_[1]);
+  my $self = $_[0]->new->activate($_[1], { always => 1 });
+  $self->build_environment_vars;
+}
+sub build_activate_environment_vars_for {
+  my $self = $_[0]->new->activate($_[1], { always => 1 });
+  $self->build_environment_vars;
+}
+sub build_deactivate_environment_vars_for {
+  my $self = $_[0]->new->deactivate($_[1]);
+  $self->build_environment_vars;
+}
+sub build_deact_all_environment_vars_for {
+  my $self = $_[0]->new->deactivate_all;
   $self->build_environment_vars;
 }
 sub build_environment_vars {
@@ -351,6 +391,12 @@ sub setup_local_lib_for {
 
 sub setup_local_lib {
   my $self = shift;
+
+  # if Carp is already loaded, ensure Carp::Heavy is also loaded, to avoid
+  # $VERSION mismatch errors (Carp::Heavy loads Carp, so we do not need to
+  # check in the other direction)
+  require Carp::Heavy if $INC{'Carp.pm'};
+
   $self->setup_env_hash;
   @INC = @{$self->inc};
 }
@@ -377,7 +423,7 @@ sub print_environment_vars_for {
 }
 
 sub environment_vars_string_for {
-  my $self = $_[0]->new->activate($_[1]);
+  my $self = $_[0]->new->activate($_[1], { always => 1});
   $self->environment_vars_string;
 }
 sub environment_vars_string {
@@ -385,8 +431,6 @@ sub environment_vars_string {
 
   $shelltype ||= $self->guess_shelltype;
 
-  my $build_method = "build_${shelltype}_env_declaration";
-
   my $extra = $self->extra;
   my @envs = (
     PATH                => $self->bins,
@@ -394,6 +438,15 @@ sub environment_vars_string {
     PERL_LOCAL_LIB_ROOT => $self->roots,
     map { $_ => $extra->{$_} } sort keys %$extra,
   );
+  $self->_build_env_string($shelltype, \@envs);
+}
+
+sub _build_env_string {
+  my ($self, $shelltype, $envs) = @_;
+  my @envs = @$envs;
+
+  my $build_method = "build_${shelltype}_env_declaration";
+
   my $out = '';
   while (@envs) {
     my ($name, $value) = (shift(@envs), shift(@envs));
@@ -405,14 +458,6 @@ sub environment_vars_string {
         && ${$value->[0]} eq $name) {
       next;
     }
-    if (
-        !ref $value
-        and defined $value
-          ? (defined $ENV{$name} && $value eq $ENV{$name})
-          : !defined $ENV{$name}
-    ) {
-      next;
-    }
     $out .= $self->$build_method($name, $value);
   }
   my $wrap_method = "wrap_${shelltype}_output";
@@ -424,21 +469,21 @@ sub environment_vars_string {
 
 sub build_bourne_env_declaration {
   my ($class, $name, $args) = @_;
-  my $value = $class->_interpolate($args, '$%s', '"', '\\%s');
+  my $value = $class->_interpolate($args, '${%s}', qr/["\\\$!`]/, '\\%s');
 
   if (!defined $value) {
     return qq{unset $name;\n};
   }
 
-  $value =~ s/(^|\G|$_path_sep)\$$name$_path_sep/$1\$$name\${$name+$_path_sep}/g;
-  $value =~ s/$_path_sep\$$name$/\${$name+$_path_sep}\$$name/;
+  $value =~ s/(^|\G|$_path_sep)\$\{$name\}$_path_sep/$1\${$name}\${$name+$_path_sep}/g;
+  $value =~ s/$_path_sep\$\{$name\}$/\${$name+$_path_sep}\${$name}/;
 
   qq{${name}="$value"; export ${name};\n}
 }
 
 sub build_csh_env_declaration {
   my ($class, $name, $args) = @_;
-  my ($value, @vars) = $class->_interpolate($args, '$%s', '"', '"\\%s"');
+  my ($value, @vars) = $class->_interpolate($args, '${%s}', '"', '"\\%s"');
   if (!defined $value) {
     return qq{unsetenv $name;\n};
   }
@@ -449,9 +494,9 @@ sub build_csh_env_declaration {
   }
 
   my $value_without = $value;
-  if ($value_without =~ s/(?:^|$_path_sep)\$$name(?:$_path_sep|$)//g) {
-    $out .= qq{if "\$$name" != '' setenv $name "$value";\n};
-    $out .= qq{if "\$$name" == '' };
+  if ($value_without =~ s/(?:^|$_path_sep)\$\{$name\}(?:$_path_sep|$)//g) {
+    $out .= qq{if "\${$name}" != '' setenv $name "$value";\n};
+    $out .= qq{if "\${$name}" == '' };
   }
   $out .= qq{setenv $name "$value_without";\n};
   return $out;
@@ -459,7 +504,7 @@ sub build_csh_env_declaration {
 
 sub build_cmd_env_declaration {
   my ($class, $name, $args) = @_;
-  my $value = $class->_interpolate($args, '%%%s%%', qr([()!^"<>&|]), '^%s');
+  my $value = $class->_interpolate($args, '%%%s%%', qr(%), '%s');
   if (!$value) {
     return qq{\@set $name=\n};
   }
@@ -467,10 +512,10 @@ sub build_cmd_env_declaration {
   my $out = '';
   my $value_without = $value;
   if ($value_without =~ s/(?:^|$_path_sep)%$name%(?:$_path_sep|$)//g) {
-    $out .= qq{\@if not "%$name%"=="" set $name=$value\n};
+    $out .= qq{\@if not "%$name%"=="" set "$name=$value"\n};
     $out .= qq{\@if "%$name%"=="" };
   }
-  $out .= qq{\@set $name=$value_without\n};
+  $out .= qq{\@set "$name=$value_without"\n};
   return $out;
 }
 
@@ -495,7 +540,7 @@ sub wrap_powershell_output {
 
 sub build_fish_env_declaration {
   my ($class, $name, $args) = @_;
-  my $value = $class->_interpolate($args, '$%s', '"', '\\%s');
+  my $value = $class->_interpolate($args, '$%s', qr/[\\"' ]/, '\\%s');
   if (!defined $value) {
     return qq{set -e $name;\n};
   }
@@ -573,7 +618,7 @@ sub resolve_home_path {
     }
   };
   unless (defined $homedir) {
-    require Carp;
+    require Carp; require Carp::Heavy;
     Carp::croak(
       "Couldn't resolve homedir for "
       .(defined $user ? $user : 'current user')
@@ -589,17 +634,26 @@ sub resolve_relative_path {
 }
 
 sub ensure_dir_structure_for {
-  my ($class, $path) = @_;
-  unless (-d $path) {
-    warn "Attempting to create directory ${path}\n";
+  my ($class, $path, $opts) = @_;
+  $opts ||= {};
+  my %dirs;
+  my @dirs = grep ! -d, (
+    $class->lib_paths_for($path),
+    $class->install_base_bin_path($path),
+  );
+  @dirs{@dirs} = ();
+  for my $dir (@dirs) {
+    require File::Basename;
+    my $parent = File::Basename::dirname($dir);
+    unless (exists $dirs{$parent} || -d $parent) {
+      $dirs{$parent} = undef;
+      push @dirs, $parent
+    }
   }
-  require File::Basename;
-  my @dirs;
-  while(!-d $path) {
-    push @dirs, $path;
-    $path = File::Basename::dirname($path);
-  }
-  mkdir $_ for reverse @dirs;
+  warn "Attempting to create directory ${path}\n"
+    if !$opts->{quiet} && @dirs;
+  mkdir $_ or die "Unable to create $_: $!"
+    for reverse @dirs;
   return;
 }
 
@@ -661,9 +715,9 @@ From the shell -
   PATH="/home/username/perl5/bin:$PATH"; export PATH;
   PERL_LOCAL_LIB_ROOT="/home/usename/perl5:$PERL_LOCAL_LIB_ROOT"; export PERL_LOCAL_LIB_ROOT;
 
-From a .bashrc file -
+From a F<.bash_profile> or F<.bashrc> file -
 
-  [ $SHLVL -eq 1 ] && eval "$(perl -I$HOME/perl5/lib/perl5 -Mlocal::lib)"
+  eval "$(perl -I$HOME/perl5/lib/perl5 -Mlocal::lib)"
 
 =head2 The bootstrapping technique
 
@@ -717,7 +771,7 @@ starts using our newly generated lib/ directory. If you are using bash or
 any other Bourne shells, you can add this to your shell startup script this
 way:
 
-  echo '[ $SHLVL -eq 1 ] && eval "$(perl -I$HOME/perl5/lib/perl5 -Mlocal::lib)"' >>~/.bashrc
+  echo 'eval "$(perl -I$HOME/perl5/lib/perl5 -Mlocal::lib)"' >>~/.bashrc
 
 If you are using C shell, you can do this as follows:
 
@@ -730,7 +784,7 @@ If you passed to bootstrap a directory other than default, you also need to
 give that as import parameter to the call of the local::lib module like this
 way:
 
-  echo '[ $SHLVL -eq 1 ] && eval "$(perl -I$HOME/foo/lib/perl5 -Mlocal::lib=$HOME/foo)"' >>~/.bashrc
+  echo 'eval "$(perl -I$HOME/foo/lib/perl5 -Mlocal::lib=$HOME/foo)"' >>~/.bashrc
 
 After writing your shell configuration file, be sure to re-read it to get the
 changed settings into your current shell's environment. Bourne shells use
@@ -757,11 +811,13 @@ installation to install modules in different directories directly this way:
   cd ../mydir2
   ... REPEAT ...
 
-When used in a C<.bashrc> file, it is recommended that you protect against
-re-activating a directory in a sub-shell.  This can be done by checking the
-C<$SHLVL> variable as shown in synopsis.  Without this, sub-shells created by
-the user or other programs will override changes made to the parent shell's
-environment.
+If you use F<.bashrc> to activate a local::lib automatically, the local::lib
+will be re-enabled in any sub-shells used, overriding adjustments you may have
+made in the parent shell.  To avoid this, you can initialize the local::lib in
+F<.bash_profile> rather than F<.bashrc>, or protect the local::lib invocation
+with a C<$SHLVL> check:
+
+  [ $SHLVL -eq 1 ] && eval "$(perl -I$HOME/perl5/lib/perl5 -Mlocal::lib)"
 
 If you are working with several C<local::lib> environments, you may want to
 remove some of them from the current environment without disturbing the others.
@@ -929,8 +985,8 @@ likely to cause issues on Win32 systems.
 
 =back
 
-Attempts to create the given path, and all required parent directories. Throws
-an exception on failure.
+Attempts to create a local::lib directory, including subdirectories and all
+required parent directories. Throws an exception on failure.
 
 =head2 print_environment_vars_for
 
